"""
Exposes KPIs API.
"""
import time
import json
import os.path
import logging
import flask
from jsonschema import validate
from flask import Blueprint
from flask import current_app as app
from app.netedit.events import TimeSlot, VehicleEvents
from app.netedit.events import EmissionEvents
from app import utils

kpis = Blueprint("kpis", __name__)

# docker
ts_url = app.dotenv["TS_URL"]
data_dir = app.dotenv["DATA_DIR"]
city = app.dotenv["URBANITE_CITY"]
assets_dir = app.dotenv["ASSETS_DIR"]

logger = logging.getLogger(__name__)

# DEBUG
@app.route("/dss/test/<int:simulation_id>", methods=["GET"])
def test_kpi_congestions(simulation_id):
    # rush_hour = utils.get_max_traffic_time(simulation_id)
    network, events = utils.get_network_and_events(simulation_id)
    ll = network.nearby_links_range("1691828242", 500)
    app.logger.debug(ll)
    result = events.share_bicycles(local_links=ll)
    return {"res": result}

@app.route("/dss/kpis/<city_id>/<int:simulation_id>", methods=["GET"])
def get_kpis_for_city(city_id, simulation_id):
    """
    Exposes API endpoint `/dss/kpis/<city_id>/<simulation_id>`-
    Returns KPIs object and stores it to storage.
    """
    app.logger.info(
        f"got calculate kpis request: city {city_id}, simulation {simulation_id}"
    )
    start_time = time.time()

    app.logger.info(f"loading networks and events")
    network, events = utils.get_network_and_events(simulation_id)
    app.logger.info(
        f"network and events loaded exec_time: {time.time()-start_time} sec"
    )

    results = {}
    if city_id == "bilbao":
        emission_object = get_emissions(simulation_id, network)
        rush_hour_time = utils.get_max_traffic_time(simulation_id)  # rush hour time
        results = prepare_results_for_bilbao(
            simulation_id, network, events, emission_object, rush_hour_time
        )
    elif city_id == "helsinki":
        emission_object = get_emissions(simulation_id, network)
        rush_hour_time = utils.get_max_traffic_time(simulation_id)  # rush hour time

        results = prepare_results_for_helsinki(
            simulation_id, network, events, emission_object, rush_hour_time
        )
    elif city_id == "messina":
        results = prepare_results_for_messina(simulation_id, network, events)
    elif city_id == "amsterdam":
        results = prepare_results_for_amsterdam(simulation_id, network, events)
    else:
        app.logger.error("get_kpis_for_city got illegal city: %s", city_id)
        raise ValueError("illegal city passed to get_kpis_for_city")
    app.logger.info(f"prepared for city")

    # validate json against schema
    schemaPath = f"{assets_dir}/schemas/kpis_{city_id}.schema.json"
    with open(schemaPath, "r", encoding="utf-8") as fd:
        schema = json.load(fd)
    validate(results, schema)
    app.logger.info(f"validated")

    # store to simulation results folder
    timestamp_dir = utils.get_sim_dir_names([simulation_id])[0]
    results_path = f"{data_dir}/{city}/simulations/{simulation_id}/results/{timestamp_dir}/kpis.json"
    app.logger.info(f"writing kpis to: {results_path}")
    with open(results_path, "w", encoding="utf-8") as fd:
        json.dump(results, fd, indent=2)

    app.logger.info("Finsihed getting kpis for the city")
    return results


@app.route("/dss/kpis/hourly/<city_id>/<int:simulation_id>", methods=["GET"])
def get_hourly_kpis_for_city(city_id, simulation_id):
    """
    Exposes API endpoint `/dss/kpis/hourly/<city_id>/<simulation_id>`-
    Returns KPIs object and stores it to storage.
    """
    app.logger.info(
        f"got calculate hourly kpis request: city {city_id}, simulation {simulation_id}"
    )
    start_time = time.time()
    timestamp_dir = utils.get_sim_dir_names([simulation_id])[0]

    # Load cached data if it exists
    results_path = f"{data_dir}/{city}/simulations/{simulation_id}/results/{timestamp_dir}/kpis_hourly.json"

    if os.path.exists(results_path):
        with open(results_path, "r", encoding="utf-8") as fp:
            kpis = json.load(fp)
            return kpis

    # Load network events
    app.logger.info(f"loading networks and events")
    network, events = utils.get_network_and_events(simulation_id)
    app.logger.info(
        f"network and events loaded exec_time: {time.time()-start_time} sec"
    )

    # Calculate per hour KPIs
    response = {}
    for slot_index in range(0, 24, 1):
        results = {}

        if city_id == "bilbao":
            emission_object = get_emissions(simulation_id, network)
            rush_hour_time = utils.get_max_traffic_time(simulation_id)  # rush hour time

            results = prepare_results_for_bilbao(
                simulation_id,
                network,
                events,
                emission_object,
                rush_hour_time,
                TimeSlot(delta=3600, slot_index=slot_index),
            )
        elif city_id == "helsinki":
            emission_object = get_emissions(simulation_id, network)
            rush_hour_time = utils.get_max_traffic_time(simulation_id)  # rush hour time

            results = prepare_results_for_helsinki(
                simulation_id,
                network,
                events,
                emission_object,
                rush_hour_time,
                TimeSlot(delta=3600, slot_index=slot_index),
            )
        elif city_id == "messina":
            results = prepare_results_for_messina(
                simulation_id,
                network,
                events,
                TimeSlot(delta=3600, slot_index=slot_index),
            )
        elif city_id == "amsterdam":
            results = prepare_results_for_amsterdam(
                simulation_id,
                network,
                events,
                TimeSlot(delta=3600, slot_index=slot_index),
            )
        else:
            app.logger.error("get_hourly_kpis_for_city got illegal city: %s", city_id)
            raise ValueError("illegal city passed to get_hourly_kpis_for_city")

        utils.deep_merge_dicts(response, results)

    # After merging the dict, simulation ID is repeated as array
    response["simulation_id"] = simulation_id

    app.logger.info(f"writing kpis to: {results_path}")
    with open(results_path, "w", encoding="utf-8") as fd:
        json.dump(response, fd, indent=2)

    return response


@app.route(
    "/dss/kpis/<city_id>/<int:simulation_id>/<locality>/<kpi_name>", methods=["GET"]
)
def return_single_kpi(city_id, simulation_id, locality, kpi_name):

    kpis_path = f"{data_dir}/{city_id}/simulations/{simulation_id}/results/{utils.get_sim_dir_names([simulation_id])[0]}/kpis.json"
    if os.path.exists(kpis_path):
        with open(kpis_path, "r", encoding="utf-8") as fp:
            kpis = json.load(fp)
            json_path = f"$..{locality}..{kpi_name}"

    rush_hour = utils.get_max_traffic_time(simulation_id)

    # kpis.json not found, calculate the KPI
    network, events = utils.get_network_and_events(simulation_id)
    if (
        kpi_name == "emissions_total"
        or kpi_name == "total_emissions_by_link"
        or kpi_name == "emissions_total_links_sum"
    ):
        emission_object = get_emissions(simulation_id, network)
        if kpi_name == "emissions_total":
            return emission_object.emissions_total()
        elif kpi_name == "total_emissions_by_link":
            return emission_object.total_emissions_by_link()
        elif kpi_name == "emissions_total_links_sum":
            return emission_object.emissions_total_links_sum()

    if kpi_name == "bike_safety_index":
        return flask.jsonify(events.bike_safety_index())
    if kpi_name == "bike_safety_aggregate":     # TODO Arguments missing
        return events.bike_safety_aggregate(events.bike_safety_index(timeslot=rush_hour))
    if kpi_name == "bikeability_aggregate":
        return events.bikeability_aggregate(events.bikeability_index())
    if kpi_name == "bikeability_index":
        return events.bikeability_index()
    if kpi_name == "vehicles_count_per_link":
        return events.vehicles_count_per_link()
    if kpi_name == "vehicles_count":
        return events.vehicles_count()
    if kpi_name == "average_bus_speed":
        return flask.jsonify(events.average_bus_speed())
    if kpi_name == "pedestrian_travel_time":
        return flask.jsonify(events.pedestrian_travel_time())
    if kpi_name == "users_per_mode":
        return events.users_per_mode()
    if kpi_name == "get_congested_links":
        return events.get_congested_links(rush_hour)["congestion_length"]
    if kpi_name == "bicycle_use":
        return flask.jsonify(events.bicycle_use())
    if kpi_name == "share_public_transport":
        return flask.jsonify(events.share_public_transport())


def get_emissions(simulation_id, network):
    """
    Returns the emissions object.
    """
    date = utils.get_sim_dir_names([simulation_id])[0]
    emissions_path_full = f"{data_dir}/{city}/simulations/{simulation_id}/results/{date}/emissionEvents.xml"
    emissions_path_short = f"{data_dir}/{city}/simulations/{simulation_id}/results/{date}/emissionEventsShort.xml"
    if not os.path.exists(emissions_path_short):
        if not os.path.exists(emissions_path_full):
            logger.warn("Path does not exist:%s", emissions_path_full)
            return None
        utils.getEventsFromKeywords(
            emissions_path_full,
            emissions_path_short,
            ["warmEmissionEvent", "coldEmissionEvent"],
        )
    emissions = EmissionEvents(path=emissions_path_short, network=network)
    return emissions


def prepare_results_for_bilbao(
    simulation_id, network, events, emisson_object, rush_hour_time, timeslot=None
):
    """
    Creates the kpis JSON object for Bilbao.
    """
    app.logger.info("prepare_results_for_bilbao with timeslot %s", timeslot)
    results = {"simulation_id": simulation_id}
    results["bilbao"] = {
        "cityWide": {"traffic": {}, "pollution": {}},
        "local": {"traffic": {}, "pollution": {}},
        "entryCapacityToCenter": {},
    }

    # cityWide
    if emisson_object != None:
        kpi = emisson_object.emissions_total(timeslot)
        results["bilbao"]["cityWide"]["pollution"] = kpi

    results["bilbao"]["cityWide"]["traffic"] = {}
    kpi = events.pedestrian_travel_time(timeslot=timeslot)
    results["bilbao"]["cityWide"]["traffic"]["pedestrianTravelTime"] = kpi["average_trip_duration"]
    # app.logger.info(kpi[0])

    kpi = events.vehicles_count(timeslot)
    results["bilbao"]["cityWide"]["traffic"]["dailyInternalBikeTravels"] = kpi["bike_count"]
    #
    # # local
    # # moyua center position in EPSG 32630
    moyua_point = {"x": 505277.192210581, "y": 4790023.046986237}
    local_links = network.nearby_links_range(moyua_point, delta=200)

    if emisson_object != None:
        kpi = emisson_object.emissions_total_links_sum(local_links, timeslot)
        results["bilbao"]["local"]["pollution"] = kpi

    results["bilbao"]["local"]["traffic"] = {}
    # kpi = events.pedestrian_travel_time(local_links, timeslot)
    # makes no sense, can't have long trips in a small area
    results["bilbao"]["local"]["traffic"]["pedestrianTravelTime"] = 0

    kpi = events.vehicles_count(local_links, timeslot)
    results["bilbao"]["local"]["traffic"]["dailyInternalBikeTravels"] = kpi[
        "bike_count"
    ]

    # If timeslot is not given, we are only interested in rush_hour_time
    if not timeslot:
        timeslot = TimeSlot(delta=3600, slot_index=rush_hour_time)

    total_capacity = events.capacity_to_moyua(local_links, timeslot)
    results["bilbao"]["entryCapacityToCenter"] = total_capacity

    return results


def prepare_results_for_amsterdam(simulation_id, network, events, timeslot=None):
    """
    Prepare kpis JSON object for Amsterdam.
    """
    rush_hour = utils.get_max_traffic_time(simulation_id)

    results = {"simulation_id": simulation_id}
    results["amsterdam"] = {
        "cityWide": {"traffic": {}, "bikeInfrastructure": {}},
        "local": {"traffic": {}, "bikeInfrastructure": {}},
    }

    # cityWide
    app.logger.info("network and events loaded - cityWide")

    kpi = events.vehicles_count(timeslot)["bike_count"]
    # Bike intensity here is interpreted as number of bike trips in the city, check vehicles_count() code for more info.
    if kpi == 0:
        kpi = 1  # TODO this needs to be fixes, as above
    results["amsterdam"]["cityWide"]["traffic"]["bikeIntensity"] = kpi
    # app.logger.warn("bike intensity: %s", kpi)

    kpi = events.get_congested_links(rush_hour, vehicle_mode="bicycle")["congestion_length"]
    results["amsterdam"]["cityWide"]["traffic"]["bikeCongestion"] = kpi
    # app.logger.warn("bike congestion: %s", kpi)

    kpi = events.bike_safety_aggregate(events.bike_safety_index(timeslot))
    results["amsterdam"]["cityWide"]["bikeInfrastructure"]["bikeSafety"] = kpi
    # app.logger.warn("bike infrastructure: %s", kpi)

    kpi = events.bikeability_aggregate(events.bikeability_index())
    results["amsterdam"]["cityWide"]["bikeInfrastructure"]["bikeability"] = kpi
    # app.logger.warn("bikeability: %s", kpi)

    # local
    app.logger.info("local kpis...")
    # gotten from EPSG:32631
    amsterdam_noord_point = {"x": 630956.8738286367, "y": 5807105.03419907}
    # a delta of 1000+ proved to be too slow
    local_links = network.nearby_links_range(amsterdam_noord_point, delta=500)

    kpi = events.vehicles_count(local_links=local_links, timeslot=timeslot)["bike_count"]
    results["amsterdam"]["local"]["traffic"]["bikeIntensity"] = kpi
    # app.logger.warn("bikeintensity: %s", kpi)

    kpi = events.get_congested_links(rush_hour, vehicle_mode="bicycle", local_links=local_links)["congestion_length"]
    results["amsterdam"]["local"]["traffic"]["bikeCongestion"] = kpi
    # app.logger.info("bike congestion: %s", kpi)

    kpi = events.bike_safety_aggregate(events.bike_safety_index(timeslot))
    results["amsterdam"]["local"]["bikeInfrastructure"]["bikeSafety"] = kpi
    # app.logger.info("bike safety: %s", kpi)

    kpi = events.bikeability_aggregate(events.bikeability_index())
    results["amsterdam"]["local"]["bikeInfrastructure"]["bikeability"] = kpi
    # app.logger.info("bikeability: %s", kpi)

    return results


def prepare_results_for_helsinki(
    simulation_id, network, events, emission_object, rush_hour_time, timeslot=None
):
    """
    Prepares the kpis JSON object for Helsinki.
    """
    results = {}
    results["simulation_id"] = simulation_id
    results["helsinki"] = {
        "cityWide": {
            "pollution": {"acousticPollution": 3},
            "congestionsAndBottlenecks": -1,
        },
        "local": {
            "pollution": {"acousticPollution": 3},
            "congestionsAndBottlenecks": -1,
        },
        "harbourAreaTrafficFlow": -1,
    }

    # cityWide
    if emission_object != None:
        kpi = emission_object.emissions_total(timeslot)
        results["helsinki"]["cityWide"]["pollution"]["emissions"] = kpi

    results["helsinki"]["cityWide"][
        "congestionsAndBottlenecks"
    ] = events.get_congested_links(rush_hour_time)["congestion_length"]

    # local
    local_links = network.nearby_links_range("4449417566", delta=500)

    if emission_object != None:
        kpi = emission_object.emissions_total_links_sum(local_links, timeslot)
        # app.logger.info(f"EMISSIONS: {kpi}")
        results["helsinki"]["local"]["pollution"] = kpi

        kpi = emission_object.total_emissions_by_link(local_links, timeslot)
        results["helsinki"]["local"]["pollutionByLink"] = kpi

    # kpi = events.get_congested_links(local_links_ids=local_links)
    results["helsinki"]["local"][
        "congestionsAndBottlenecks"
    ] = events.get_congested_links(rush_hour_time, local_links=local_links)["congestion_length"]

    return results


def prepare_results_for_messina(simulation_id, network, events, timeslot=None):
    """
    Prepares the kpis JSON object for Messina.
    """
    app.logger.info("prepare_results_for_messina with timeslot %s", timeslot)
    results = {}
    results["simulation_id"] = simulation_id
    results["messina"] = {"cityWide": {}, "local": {}}

    # cityWide
    app.logger.info("network and events loaded")
    vehicles_count = events.vehicles_count(timeslot)

    kpi = events.average_bus_speed(timeslot)
    results["messina"]["cityWide"]["publicTransport"] = {}
    results["messina"]["cityWide"]["publicTransport"][
        "averageSpeedOfPublicTransport"
    ] = kpi
    results["messina"]["cityWide"]["publicTransport"][
        "numberOfBikeTrips"
    ] = vehicles_count["bike_count"]
    kpi = events.users_per_mode(timeslot)
    results["messina"]["cityWide"]["publicTransport"]["publicTransportUse"] = kpi["PT"]

    results["messina"]["cityWide"]["shareOfTrips"] = {}
    kpi = events.share_public_transport(timeslot)
    results["messina"]["cityWide"]["shareOfTrips"]["shareOfPublicTransport"] = kpi
    kpi = events.share_cars(timeslot)
    results["messina"]["cityWide"]["shareOfTrips"]["shareOfCars"] = kpi
    kpi = events.share_bicycles(timeslot)
    results["messina"]["cityWide"]["shareOfTrips"]["shareOfBicycles"] = kpi

    # Local
    results["messina"]["local"]["publicTransport"] = {}

    local_links = network.nearby_links_range("1691828242", delta=1800)
    vehicles_count = events.vehicles_count(local_links=local_links, timeslot=timeslot)
    kpi = events.average_bus_speed(local_links=local_links, timeslot=timeslot)
    results["messina"]["local"]["publicTransport"][
        "averageSpeedOfPublicTransport"
    ] = kpi
    results["messina"]["local"]["publicTransport"][
        "numberOfBikeTrips"
    ] = vehicles_count["bike_count"]
    kpi = events.users_per_mode(local_links=local_links, timeslot=timeslot)
    results["messina"]["local"]["publicTransport"]["publicTransportUse"] = kpi["PT"]

    results["messina"]["local"]["shareOfTrips"] = {}
    kpi = events.share_public_transport(local_links=local_links, timeslot=timeslot)
    results["messina"]["local"]["shareOfTrips"]["shareOfPublicTransport"] = kpi
    kpi = events.share_cars(local_links=local_links, timeslot=timeslot)
    results["messina"]["local"]["shareOfTrips"]["shareOfCars"] = kpi
    kpi = events.share_bicycles(local_links=local_links, timeslot=timeslot)
    results["messina"]["local"]["shareOfTrips"]["shareOfBicycles"] = kpi

    return results
